(in-package :pila)
(export 'main)
(export 'rpn)

(defparameter +prelude-path+ "prelude.pila")

(defmethod print-object ((self word) stream)
  (print-unreadable-object (self stream :type t)
    (format t "~s" (word-name self))))

(defgeneric eval-word (self)
  (:documentation "Evaluates a word, whether is a native or user-defined word."))

(defmethod eval-word ((self native-word))
  (funcall (fn-ptr self)))

(defmethod eval-word ((num number))
  (vector-push-extend num *rpn-stack*))

(defmethod eval-word ((v boolean-word)) v)

(defmethod eval-word ((self user-word))
  (loop for word-def across (user-word-def-array self) do
        (eval-word word-def)))

(defmacro is-quoted-word (word)
 `(and (> (length ,word) 1) (char= (schar ,word 0) #\')))

(defun get-word (word)
  (declare (optimize (speed 3))
           (type string word))
  (let ((is-quoted (is-quoted-word word))
        (org-word word))
    (when is-quoted
      (setf word (string-slice word 1)))
    (multiple-value-bind (word-impl exists) (gethash word *rpn-dict*)
      (unless exists
        (error "no such word `~a` in the dictionary" word))
      (if is-quoted
          (make-native-word
            org-word
            (lambda () (vector-push-extend word-impl *rpn-stack*)))
          word-impl))))

(defun dispatch-word (word)
  (declare (optimize (speed 3)))
  (declare (type string word))
  (if (is-quoted-word word)
     (vector-push-extend (get-word (string-slice word 1)) *rpn-stack*)
     (eval-word (get-word word))))

(defun rpn-eval (prog)
  (declare (optimize (speed 3)))
  (dolist (word (str:words prog))
    (declare (type (simple-array character (*)) word))
    (cond
      ((string= word "\\")
       (return))
      ((and *compile-mode* (null *current-def-word*))
       (setf *current-def-word* (make-user-word word (make-dynarray 64)))
       (setf (gethash word *rpn-dict*) *current-def-word*))
      ((is-number-p (schar word 0))
       (let ((num (parse-num word)))
         (if *compile-mode*
             (vector-push-extend num (user-word-def-array *current-def-word*))
             (vector-push-extend num *rpn-stack*))))
      (t ; a word
        (if (and *compile-mode* (string/= word ";"))
            (vector-push-extend (get-word word)
                         (user-word-def-array *current-def-word*))
            (dispatch-word word))))))

(defmacro rpn-repl-handle-errors (entry)
  `(flet ((clear-compile ()
            (setf *compile-mode* nil *current-def-word* nil)))
     (handler-case ,entry
       (simple-error (c)
         (format t "error at ~d: ~a~%" *line-number* c)
         (clear-compile))
       (simple-type-error (c)
         (format t "type error: ~a~%" c)
         (clear-compile)))))

(defun read-and-eval-stream (stream)
  (declare (type stream stream))
  (multiple-value-bind (user-line eof?) (read-line stream nil)
    (unless eof?
      (rpn-repl-handle-errors
        (rpn-eval user-line))
      (incf *line-number*)
      t)))

(defun load-file (file-name)
  (with-open-file (stream file-name :direction :input)
    (loop while (read-and-eval-stream stream))))

(defun repl ()
  (loop
   (format t "(~d)~:[@~;:~] " *line-number* *compile-mode*)
   (finish-output)
   (when (null (read-and-eval-stream *standard-input*))
     (return))))

(defun main ()
  (write-line "Loading prelude...")
  (if (probe-file +prelude-path+)
      (load-file +prelude-path+)
      (write-line "Could not find the prelude!" *error-output*))
  (setf *line-number* 1)
  (let ((command-line (uiop:command-line-arguments)))
    (if (null command-line)
        (repl)
        (load-file (first command-line))))
  (uiop:quit))
